import pytest

from lynse.utils.utils import OpsError

import lynse
import numpy as np

client = lynse.VectorDBClient()
database = client.create_database(database_name='test_local_db', drop_if_exists=True)


def test_modify_exists_database_parameter():
    collection = database.require_collection("test_collection", dim=1024, chunk_size=10000,
                                             dtypes='float32', drop_if_exists=True)

    # insert data
    with collection.insert_session():
        for i in range(10):
            collection.add_item(np.random.rand(1024), id=i)

    collection = database.require_collection("test_collection", dim=1024, chunk_size=10000,
                                             dtypes='float32', drop_if_exists=False)

    assert collection._matrix_serializer.dim == 1024  # not modified
    assert collection._matrix_serializer.chunk_size == 10000  # not modified
    assert collection._matrix_serializer.dtypes == np.float32  # not modified


def test_using_api_after_database_deleted(capfd):
    collection = database.require_collection("test_collection2", dim=1024, chunk_size=10000, dtypes='float32', drop_if_exists=True)

    # insert data
    with collection.insert_session():
        for i in range(10):
            collection.add_item(np.random.rand(1024), id=i)

    # delete database
    collection.delete()

    # using add_item function after database deleted, it should raise ValueError
    with pytest.raises(OpsError):
        collection.add_item(np.random.rand(1024), id=11)

    # using bulk_add_items function after database deleted, it should raise ValueError
    with pytest.raises(OpsError):
        collection.bulk_add_items([(np.random.rand(1024), 11)])

    # using query function after database deleted, it should raise ValueError
    with pytest.raises(OpsError):
        collection.search(np.random.rand(1024))



from pyroaring import BitMap


class IDChecker:
    def __init__(self):
        self.ids = BitMap()

    def add(self, items):
        if isinstance(items, int):
            self.ids.add(items)
        else:
            self.ids.update(items)

    def concat(self, another_id_checker):
        if not isinstance(another_id_checker, IDChecker):
            raise ValueError("Expected another_id_checker to be an instance of IDChecker")

        self.ids.update(another_id_checker.ids)

    def drop(self, items):
        if isinstance(items, int):
            if items in self.ids:
                self.ids.discard(items)
        else:
            for item in items:
                if item in self.ids:
                    self.ids.discard(item)

    def __contains__(self, item):
        return item in self.ids

    def to_file(self, filepath):
        with open(filepath, 'wb') as file:
            file.write(self.ids.serialize())

    def from_file(self, filepath):
        try:
            with open(filepath, 'rb') as file:
                self.ids = BitMap.deserialize(file.read())
        except FileNotFoundError:
            self.ids = BitMap()

    def find_max_value(self):
        return max(self.ids) if self.ids else -1

    def filter_ids(self, matcher):
        """
        Filter the IDs based on the matcher.

        Parameters:
            matcher: Matcher
                The matcher to use for filtering.

        Returns:
            List[int]: The filtered IDs.
        """
        if matcher.matcher.name == "MatchField":
            comparator_name = matcher.matcher.comparator.__name__
            value = matcher.matcher.value

            if comparator_name == 'eq':
                return [value] if value in self.ids else []
            elif comparator_name == 'ne':
                return list(self.ids - BitMap([value]))
            elif comparator_name == 'le':
                return [id for id in self.ids if id <= value]
            elif comparator_name == 'ge':
                return [id for id in self.ids if id >= value]
            elif comparator_name == 'lt':
                return [id for id in self.ids if id < value]
            elif comparator_name == 'gt':
                return [id for id in self.ids if id > value]
        elif matcher.matcher.name == "MatchRange":
            start, end, inclusive = matcher.matcher.start, matcher.matcher.end, matcher.matcher.inclusive
            if inclusive is True:
                return [id for id in self.ids if start <= id <= end]
            elif inclusive is False:
                return [id for id in self.ids if start < id < end]
            elif inclusive == "left":
                return [id for id in self.ids if start <= id < end]
            elif inclusive == "right":
                return [id for id in self.ids if start < id <= end]

        return [id for id in self.ids if matcher.evaluate(data=None, external_id=id)]

    def retrieve_all_ids(self, return_set=False):
        return list(self.ids) if not return_set else set(self.ids)



from pathlib import Path
from typing import Callable, Union

import numpy as np
from spinesUtils.asserts import raise_if

from ..configs.config import config
from ..storage_layer.storage import PersistentFileStorage
from ..core_components.kmeans import BatchKMeans
from ..core_components.ivf_index import IVFIndex
from ..core_components.locks import ThreadLock



class InvertedIndex:
    def __init__(self):
        pass


import os
import shutil
import time
import pytest

from test import Filter, FieldCondition, MatchField, MatchID
import numpy as np
import lynse

client = lynse.VectorDBClient()
database = client.create_database(database_name='test_local_db', drop_if_exists=True)


@pytest.fixture
def cleanup():
    # 在测试前执行清理
    db_path = os.path.expanduser('~/.LynseDB/databases/test_local_db')
    if os.path.exists(db_path):
        shutil.rmtree(db_path)
    time.sleep(0.1)  # 给系统一些时间来完全释放资源

    yield  # 这里会运行测试

    # 在测试后执行清理
    if os.path.exists(db_path):
        shutil.rmtree(db_path)
    time.sleep(0.1)


def get_collection(dim=100, chunk_size=1000, dtypes='float32', drop_if_exists=True):
    collection = database.require_collection('test_collection', dim=dim, chunk_size=chunk_size,
                                             dtypes=dtypes, drop_if_exists=drop_if_exists)
    return collection


def get_collection_for_query(*args, with_field=True, field_prefix='test_', **kwargs):
    collection = get_collection(*args, **kwargs)
    np.random.seed(2023)
    with collection.insert_session():
        items = []
        for i in range(100):
            items.append((np.random.random(100), i, {"test": field_prefix + str(i // 10) if with_field else None}))

        collection.bulk_add_items(items, enable_progress_bar=False)

    return collection


def test_add_single_item_without_id_and_field(cleanup):
    collection = get_collection(drop_if_exists=True)
    id = collection.add_item(np.ones(100), id=1)

    collection.commit()

    assert collection.shape == (1, 100)


def test_add_single_item_with_id_and_field(cleanup):
    collection = get_collection(drop_if_exists=True)
    id = collection.add_item(np.ones(100), id=1, field={"test": 1})

    collection.commit()

    assert collection.shape == (1, 100)


def test_bulk_add_item_without_id_and_field(cleanup):
    collection = get_collection(drop_if_exists=True)
    items = []
    for i in range(101):
        items.append((np.ones(100), i))

    collection.bulk_add_items(items)

    collection.commit()

    assert collection.shape == (101, 100)


def test_bulk_add_item_with_id_and_field(cleanup):
    collection = get_collection(drop_if_exists=True)
    items = []
    for i in range(101):
        items.append((np.ones(100), i, {"test": "test_" + str(i // 10)}))

    collection.bulk_add_items(items)

    collection.commit()

    assert collection.shape == (101, 100)


def test_add_bulk_item_with_id_and_chinese_field(cleanup):
    collection = get_collection(drop_if_exists=True)
    items = []
    for i in range(101):
        items.append((np.ones(100), i, {"测试": "测试_" + str(i // 10)}))

    collection.bulk_add_items(items)

    collection.commit()

    assert collection.shape == (101, 100)


def test_query_without_field(cleanup):
    collection = get_collection_for_query(with_field=False)

    vec = np.random.random(100)
    n, d, f = collection.search(vec, k=6)

    assert len(n) == len(d) == 6
    assert list(d) == sorted(d, key=lambda s: -s)
    assert all(i in collection._id_filter for i in n)


def test_query_with_field(cleanup):
    collection = get_collection_for_query(with_field=True, drop_if_exists=True)

    vec = np.random.random(100)
    n, d, f = collection.search(
        vec, k=6, search_filter=Filter(
            must=[FieldCondition(key="test", matcher=MatchField('test_1'))]
        )
    )

    assert len(n) == len(d) == 6
    assert list(d) == sorted(d, key=lambda s: -s)
    assert all(i in collection._id_filter for i in n)
    assert all(10 <= i < 20 for i in n)


def test_query_with_list_field(cleanup):
    collection = get_collection_for_query(with_field=True, drop_if_exists=True)

    vec = np.random.random(100)
    n, d, f = collection.search(
        vec, k=12, search_filter=Filter(
            any=[
                FieldCondition(key="test", matcher=MatchField('test_1')),
                FieldCondition(key="test", matcher=MatchField('test_7'))
            ]
        )
    )

    assert len(n) == len(d) == 12
    assert list(d) == sorted(d, key=lambda s: -s)
    assert all(i in collection._id_filter for i in n)
    assert all((10 <= i < 20) or (70 <= i < 80) for i in n)


def test_query_with_chinese_field(cleanup):
    collection = get_collection_for_query(with_field=True, field_prefix='测试_')

    vec = np.random.random(100)
    n, d, f = collection.search(
        vec, k=6, search_filter=Filter(
            must=[
                FieldCondition(key="test", matcher=MatchField('测试_1'))
            ]
        )
    )

    assert len(n) == len(d) == 6
    assert list(d) == sorted(d, key=lambda s: -s)
    assert all(i in collection._id_filter for i in n)
    assert all(10 <= i < 20 for i in n)


def test_query_with_chinese_list_field(cleanup):
    collection = get_collection_for_query(with_field=True, field_prefix='测试_')

    vec = np.random.random(100)
    n, d, f = collection.search(
        vec, k=12, search_filter=Filter(
            any=[
                FieldCondition(key="test", matcher=MatchField('测试_1')),
                FieldCondition(key="test", matcher=MatchField('测试_7'))
            ]
        )
    )

    assert len(n) == len(d) == 12
    assert list(d) == sorted(d, key=lambda s: -s)
    assert all(i in collection._id_filter for i in n)
    assert all((10 <= i < 20) or (70 <= i < 80) for i in n)


def test_query_with_subset_indices(cleanup):
    collection = get_collection_for_query(with_field=True, drop_if_exists=True)

    vec = np.random.random(100)
    n, d, f = collection.search(
        vec, k=6, search_filter=Filter(
            must=[FieldCondition(key=":match_id:", matcher=MatchID(list(range(10))))]
        )
    )

    assert len(n) == len(d) == 6
    assert list(d) == sorted(d, key=lambda s: -s)
    assert all(i in collection._id_filter for i in n)
    assert all(i < 10 for i in n)


def test_query_with_subset_indices_and_field(cleanup):
    collection = get_collection_for_query(with_field=True, field_prefix='test_')

    vec = np.random.random(100)
    n, d, f = collection.search(
        vec, k=6, search_filter=Filter(
            must=[
                FieldCondition(key="test", matcher=MatchField('test_0')),
                FieldCondition(key=":match_id:", matcher=MatchID(list(range(10))))
            ]
        )
    )

    assert len(n) == len(d) == 6
    assert list(d) == sorted(d, key=lambda s: -s)
    assert all(i in collection._id_filter for i in n)
    assert all(i < 10 for i in n)


def test_query_with_subset_indices_and_list_field(cleanup):
    collection = get_collection_for_query(with_field=True, drop_if_exists=True)

    vec = np.random.random(100)
    n, d, f = collection.search(
        vec, k=12, search_filter=Filter(
            must=[
                FieldCondition(key=":match_id:", matcher=MatchID(list(range(10, 20)) + list(range(70, 80))))
            ],
            any=[
                FieldCondition(key="test", matcher=MatchField('test_1')),
                FieldCondition(key="test", matcher=MatchField('test_7')),
            ]
        )
    )

    assert len(n) == len(d) == 12
    assert list(d) == sorted(d, key=lambda s: -s)
    assert all(i in collection._id_filter for i in n)
    assert all((10 <= i < 20) or (70 <= i < 80) for i in n)


def test_query_with_subset_indices_and_chinese_field(cleanup):
    collection = get_collection_for_query(with_field=True, field_prefix='测试_')

    vec = np.random.random(100)
    n, d, f = collection.search(
        vec, k=6, search_filter=Filter(
            must=[
                FieldCondition(key="test", matcher=MatchField('测试_0')),
                FieldCondition(key=":match_id:", matcher=MatchID(list(range(10))))
            ]
        )
    )

    assert len(n) == len(d) == 6
    assert list(d) == sorted(d, key=lambda s: -s)
    assert all(i in collection._id_filter for i in n)
    assert all(i < 10 for i in n)


def test_query_with_subset_indices_and_chinese_list_field(cleanup):
    collection = get_collection_for_query(with_field=True, field_prefix='测试_')

    vec = np.random.random(100)
    n, d, f = collection.search(
        vec, k=12, search_filter=Filter(
            must=[
                FieldCondition(key=":match_id:", matcher=MatchID(list(range(10, 20)) + list(range(70, 80))))
            ],
            any=[
                FieldCondition(key="test", matcher=MatchField('测试_1')),
                FieldCondition(key="test", matcher=MatchField('测试_7')),
            ]
        )
    )

    print(n)
    assert len(n) == len(d) == 12
    assert list(d) == sorted(d, key=lambda s: -s)
    assert all(i in collection._id_filter for i in n)
    assert all((10 <= i < 20) or (70 <= i < 80) for i in n)


def test_query_stability_of_mvdb_files(cleanup):
    collection = get_collection_for_query(with_field=True, drop_if_exists=True)
    vec = np.random.random(100)
    last_n, last_d, last_f = collection.search(
        vec, k=6,
        search_filter=Filter(
            must=[FieldCondition(key="test", matcher=MatchField('test_1'))]
        )
    )

    assert len(last_n) == len(last_d) == 6
    for i in range(20):
        n, d, f = collection.search(
            vec, k=6,
            search_filter=Filter(
                must=[FieldCondition(key="test", matcher=MatchField('test_1'))]
            )
        )
        assert last_d.tolist() == d.tolist()
        assert last_n.tolist() == n.tolist()
        assert len(n) == len(d) == 6
        assert list(d) == sorted(d, key=lambda s: -s)
        assert all(i in collection._id_filter for i in n)
        assert all(10 <= i < 20 for i in n)


def test_multiple_bulk_add_items(cleanup):
    collection = get_collection(drop_if_exists=True)
    items = []
    for i in range(101):
        items.append((np.ones(100), i))

    collection.bulk_add_items(items)
    collection.commit()
    assert collection.shape == (101, 100)

    items = []
    for i in range(101):
        items.append((np.ones(100), i + 101))

    collection.bulk_add_items(items)
    collection.commit()

    assert collection.shape == (202, 100)


def test_multiple_bulk_add_items_with_insert_session(cleanup):
    collection = get_collection(drop_if_exists=True)
    items = []
    for i in range(101):
        items.append((np.ones(100), i))

    with collection.insert_session():
        collection.bulk_add_items(items)

    assert collection.shape == (101, 100)

    items = []
    for i in range(101, 202):
        items.append((np.ones(100), i))

    with collection.insert_session():
        collection.bulk_add_items(items)

    assert collection.shape == (202, 100)


def test_multiple_initialization(cleanup):
    dim = 100; chunk_size = 1000; dtypes = 'float32'
    collection = get_collection(drop_if_exists=True, dim=dim, chunk_size=chunk_size, dtypes=dtypes)
    items = []
    for i in range(101):
        items.append((np.ones(100), i, {"test": "test_" + str(i // 10)}))

    with collection.insert_session():
        collection.bulk_add_items(items)

    assert collection.shape == (101, 100)

    collection = get_collection(dim=dim, chunk_size=chunk_size, dtypes=dtypes, drop_if_exists=False)

    items = []
    for i in range(101):
        items.append((np.ones(100), i + 101, {"test": "test_" + str(i // 10)}))
    # insert
    with collection.insert_session():
        collection.bulk_add_items(items)

    assert collection.shape == (202, 100)


def test_result_order(cleanup):
    def get_test_vectors(shape):
        for i in range(shape[0]):
            yield np.random.random(shape[1])

    collection = get_collection(drop_if_exists=True, dim=10, chunk_size=10000)
    with collection.insert_session():
        # Define the initial ID.
        id = 0
        vectors = []
        for t in get_test_vectors((100000, 10)):
            if id == 0:
                query = t
            vectors.append((t, id))
            id += 1

        # Here, normalization can be directly specified, achieving the same effect as `t = t / np.linalg.norm(t) `.
        collection.bulk_add_items(vectors)

    for index_mode in ['IVF-IP-SQ8', 'IVF-IP', 'IVF-L2sq-SQ8', 'IVF-L2sq',
                       'IVF-Cos-SQ8', 'IVF-Cos', 'IVF-Jaccard-Binary', 'IVF-Hamming-Binary',
                       'Flat-IP-SQ8', 'Flat-IP', 'Flat-L2sq-SQ8', 'Flat-L2sq', 'Flat-Cos-SQ8', 'Flat-Cos',
                       'Flat-Jaccard-Binary', 'Flat-Hamming-Binary']:
        collection.build_index(index_mode=index_mode)
        index, score, field = collection.search(query, k=10)

        assert len(index) == len(score) == 10


def test_transactions(cleanup):
    collection = get_collection(dim=1024, chunk_size=10000, drop_if_exists=True)

    def get_test_vectors(shape):
        for i in range(shape[0]):
            yield np.random.random(shape[1])

    with collection.insert_session():
        id = 0
        vectors = []
        for t in get_test_vectors((100000, 1024)):
            vectors.append((t, id))
            id += 1

        collection.bulk_add_items(vectors, batch_size=1000)

    collection.build_index(index_mode='IVF-IP')
    assert collection.shape == (100000, 1024)

    with pytest.raises(ValueError):
        with collection.insert_session():
            id = 0
            vectors = []
            for t in get_test_vectors((100000, 1024)):
                vectors.append((t, id))
                id += 1
            collection.bulk_add_items(vectors, batch_size=1000)

    collection.build_index(index_mode='IVF-IP')
    assert collection.shape == (100000, 1024)

    with pytest.raises(ValueError):
        collection.bulk_add_items(vectors, batch_size=1000)

    collection.build_index(index_mode='IVF-IP')
    assert collection.shape == (100000, 1024)


def test_filter(cleanup):
    collection = get_collection_for_query(with_field=True, drop_if_exists=True)

    vec = np.random.random(100)
    n, d, f = collection.search(
        vec, k=6, search_filter=Filter(
            must=[FieldCondition(key="test", matcher=MatchField('test_1'))]
        )
    )

    assert len(n) == len(d) == 6
    assert list(d) == sorted(d, key=lambda s: -s)
    assert all(i in collection._id_filter for i in n)
    assert all(10 <= i < 20 for i in n)

    n, d, f = collection.search(
        vec, k=6, search_filter=Filter(
            must=[FieldCondition(key="test", matcher=MatchField('test_1'))],
            any=[FieldCondition(key="test", matcher=MatchField('test_7'))]
        )
    )

    assert len(n) == len(d) == 0

    n, d, f = collection.search(
        vec, k=6, search_filter=Filter(
            must=[FieldCondition(key="test", matcher=MatchField('test_1'))],
            any=[FieldCondition(key="test", matcher=MatchField('test_0'))],
            must_not=[FieldCondition(key="test", matcher=MatchField('test_0'))]
        )
    )

    assert len(n) == len(d) == 0

    n, d, f = collection.search(
        vec, k=6, search_filter=Filter(
            must=[FieldCondition(key=":match_id:", matcher=MatchID([1, 2, 3]))],
            any=[FieldCondition(key="test", matcher=MatchField('test_0'))],
        )
    )

    assert len(n) == len(d) == 3
    assert list(d) == sorted(d, key=lambda s: -s)
    assert all(i in collection._id_filter for i in n)
    assert all(1 <= i < 4 for i in n)

    n, d, f = collection.search(
        vec, k=6, search_filter=Filter(
            must=[FieldCondition(key=":match_id:", matcher=MatchID([1, 2, 3]))],
            any=[FieldCondition(key="test", matcher=MatchField('test_0'))],
            must_not=[FieldCondition(key="test", matcher=MatchField('test_0'))]
        )
    )

    assert len(n) == len(d) == 0

    n, d, f = collection.search(
        vec, k=6, search_filter=Filter(
            must=[FieldCondition(key=":match_id:", matcher=MatchID([1, 2, 3]))],
            any=[FieldCondition(key="test", matcher=MatchField('test_0'))],
            must_not=[FieldCondition(key=":match_id:", matcher=MatchID([1, 2, 3]))]
        )
    )

    assert len(n) == len(d) == 0

    n, d, f = collection.search(
        vec, k=6, search_filter=Filter(
            must=[FieldCondition(key=":match_id:", matcher=MatchID([1, 2, 3]))],
            any=[FieldCondition(key="test", matcher=MatchField('test_0'))],
            must_not=[FieldCondition(key=":match_id:", matcher=MatchID([4, 5, 6]))]
        )
    )

    assert len(n) == len(d) == 3
    assert list(d) == sorted(d, key=lambda s: -s)
    assert all(i in collection._id_filter for i in n)
    assert all(1 <= i < 4 for i in n)

    n, d, f = collection.search(
        vec, k=6, search_filter=Filter(
            must=[
                FieldCondition(key='test', matcher=MatchField('test_0')),
            ],
            any=[
                FieldCondition(key=":match_id:", matcher=MatchID([1, 2, 3, 4, 5])),
            ]
        ))

    assert len(n) == len(d) == 5
    assert list(d) == sorted(d, key=lambda s: -s)
    assert all(i in collection._id_filter for i in n)
    assert all(0 <= i < 6 for i in n)

    n, d, f = collection.search(
        vec, k=6, search_filter=Filter(
            must=[
                FieldCondition(key='test', matcher=MatchField(['test_0', 'test_00'], all_comparators=True)),
            ],
            any=[
                FieldCondition(key=":match_id:", matcher=MatchID([1, 2, 3, 4, 5])),
            ]
        ))

    assert len(n) == len(d) == 0

    n, d, f = collection.search(
        vec, k=6, search_filter=Filter(
            must=[
                FieldCondition(key='test', matcher=MatchField(['test_0', 'test_00'], all_comparators=False)),
            ],
            any=[
                FieldCondition(key=":match_id:", matcher=MatchID([1, 2, 3, 4, 5])),
            ]
        ))

    assert len(n) == len(d) == 5
    assert list(d) == sorted(d, key=lambda s: -s)
    assert all(i in collection._id_filter for i in n)
    assert all(0 <= i < 6 for i in n)



import os
import msgpack


class TrieNode:
    __slots__ = ['children', 'is_end_of_word', 'ids']

    def __init__(self):
        self.children = {}
        self.is_end_of_word = False
        self.ids = set()


class Trie:
    def __init__(self):
        self.root = TrieNode()

    def insert(self, key, record_id):
        node = self.root
        for char in key:
            if char not in node.children:
                node.children[char] = TrieNode()
            node = node.children[char]
        node.is_end_of_word = True
        node.ids.add(record_id)

    def search(self, key):
        if isinstance(key, (list, tuple)):
            return self._batch_search(map(str, key))
        else:
            return self._search_single(str(key))

    def _batch_search(self, keys):
        results = set()
        current_nodes = {self.root: set(keys)}

        while current_nodes:
            next_nodes = {}
            for node, prefixes in current_nodes.items():
                for prefix in prefixes:
                    if not prefix:
                        if node.is_end_of_word:
                            results.update(node.ids)
                    else:
                        char = prefix[0]
                        if char in node.children:
                            child = node.children[char]
                            next_prefix = prefix[1:]
                            if child not in next_nodes:
                                next_nodes[child] = set()
                            next_nodes[child].add(next_prefix)
            current_nodes = next_nodes

        return results

    def starts_with(self, prefix):
        def collect_all_words(node):
            stack = [node]
            ids = set()
            while stack:
                node = stack.pop()
                if node.is_end_of_word:
                    ids.update(node.ids)
                for child in node.children.values():
                    stack.append(child)
            return ids

        node = self.root
        for char in prefix:
            if char not in node.children:
                return set()
            node = node.children[char]
        return collect_all_words(node)

    def serialize(self):
        stack = [(self.root, "")]
        serialized = {}
        while stack:
            node, prefix = stack.pop()
            serialized[prefix] = {
                "children": list(node.children.keys()),
                "is_end_of_word": node.is_end_of_word,
                "ids": list(node.ids)
            }
            for char, child in node.children.items():
                stack.append((child, prefix + char))
        return serialized

    def deserialize(self, data):
        root = TrieNode()
        nodes = {"": root}
        for prefix, node_data in data.items():
            node = nodes[prefix]
            node.is_end_of_word = node_data["is_end_of_word"]
            node.ids = set(node_data["ids"])
            for char in node_data["children"]:
                child_prefix = prefix + char
                child_node = TrieNode()
                node.children[char] = child_node
                nodes[child_prefix] = child_node
        self.root = root


class BPlusTreeNode:
    __slots__ = ['is_leaf', 'keys', 'children']

    def __init__(self, is_leaf=False):
        self.is_leaf = is_leaf
        self.keys = []
        self.children = []


class BPlusTree:
    def __init__(self, t=3):
        self.root = BPlusTreeNode(is_leaf=True)
        self.t = t

    def insert(self, key, record_id):
        root = self.root
        if len(root.keys) == (2 * self.t - 1):
            temp = BPlusTreeNode()
            self.root = temp
            temp.children.insert(0, root)
            self._split_child(temp, 0)
        self._insert_non_full(self.root, key, record_id)

    def _insert_non_full(self, node, key, record_id):
        if node.is_leaf:
            idx = len(node.keys) - 1
            while idx >= 0 and key < node.keys[idx][0]:
                idx -= 1
            node.keys.insert(idx + 1, (key, record_id))
        else:
            idx = len(node.keys) - 1
            while idx >= 0 and key < node.keys[idx][0]:
                idx -= 1
            idx += 1
            if len(node.children[idx].keys) == (2 * self.t - 1):
                self._split_child(node, idx)
                if key > node.keys[idx][0]:
                    idx += 1
            self._insert_non_full(node.children[idx], key, record_id)

    def _split_child(self, parent, idx):
        t = self.t
        child = parent.children[idx]
        new_node = BPlusTreeNode(is_leaf=child.is_leaf)
        parent.keys.insert(idx, child.keys[t - 1])
        parent.children.insert(idx + 1, new_node)

        new_node.keys = child.keys[t:(2 * t - 1)]
        child.keys = child.keys[0:(t - 1)]

        if not child.is_leaf:
            new_node.children = child.children[t:(2 * t)]
            child.children = child.children[0:t]

    def search(self, key):
        return self._search(self.root, key)

    def _search(self, node, key):
        idx = 0
        while idx < len(node.keys) and key > node.keys[idx][0]:
            idx += 1
        if idx < len(node.keys) and key == node.keys[idx][0]:
            return {node.keys[idx][1]}
        if node.is_leaf:
            return set()
        return self._search(node.children[idx], key)

    def range_search(self, start_key, end_key):
        result = set()
        self._range_search(self.root, start_key, end_key, result)
        return result

    def _range_search(self, node, start_key, end_key, result):
        idx = 0
        while idx < len(node.keys) and (start_key is not None and node.keys[idx][0] < start_key):
            idx += 1

        if node.is_leaf:
            while idx < len(node.keys) and (end_key is None or node.keys[idx][0] <= end_key):
                result.add(node.keys[idx][1])
                idx += 1
        else:
            while idx < len(node.keys) and (end_key is None or node.keys[idx][0] <= end_key):
                self._range_search(node.children[idx], start_key, end_key, result)
                result.add(node.keys[idx][1])
                idx += 1
            if idx < len(node.children):
                self._range_search(node.children[idx], start_key, end_key, result)


class Index:
    def __init__(self):
        self.indices = {}
        self.index_schema = {}

    def add_index(self, field, field_type, rebuild_if_exists=False):
        if field_type not in [int, float, str]:
            raise ValueError("Unsupported field type. Supported types are int, float, and str.")
        if field in self.indices:
            if rebuild_if_exists:
                self.remove_index(field)
            else:
                return

        if field_type == str:
            self.indices[field] = Trie()
        else:
            self.indices[field] = BPlusTree()

        self.index_schema[field] = field_type

    def insert(self, record, record_id):
        for field, index in self.indices.items():
            field_value = record.get(field)
            if field_value is not None:
                index.insert(field_value, record_id)

    def search(self, field, value):
        if field not in self.indices:
            raise ValueError(f"No index for field '{field}'")
        return self.indices[field].search(value)

    def range_search(self, field, start_value, end_value):
        if field not in self.indices:
            raise ValueError(f"No index for field '{field}'")
        if isinstance(self.indices[field], Trie):
            raise ValueError("Range search is not supported for Trie index.")
        return self.indices[field].range_search(start_value, end_value)

    def save(self, filepath):
        with open(filepath, 'wb') as f:
            packed_indices = msgpack.packb((self._serialize_indices(),
                                            {k: str(v.__name__) for k, v in self.index_schema.items()}),
                                           use_bin_type=True)
            f.write(packed_indices)

    def load(self, filepath):
        if os.path.exists(filepath):
            with open(filepath, 'rb') as f:
                packed_indices = f.read()
                indices, index_schema = msgpack.unpackb(packed_indices, raw=False)
                self._deserialize_indices(indices)
                self.index_schema = {k: eval(v) for k, v in index_schema.items()}
        return self

    def _serialize_indices(self):
        serialized = {}
        for field, index in self.indices.items():
            if isinstance(index, Trie):
                serialized[field] = {
                    "type": "trie",
                    "data": index.serialize()
                }
            elif isinstance(index, BPlusTree):
                serialized[field] = {
                    "type": "bplustree",
                    "data": self._serialize_bplustree(index)
                }
        return serialized

    def _deserialize_indices(self, data):
        for field, index_data in data.items():
            if index_data["type"] == "trie":
                index = Trie()
                index.deserialize(index_data["data"])
            else:
                index = BPlusTree()
                self._deserialize_bplustree(index, index_data["data"])
            self.indices[field] = index

    @staticmethod
    def _serialize_bplustree(bplustree):
        serialized_nodes = []

        def serialize_node(node):
            node_data = {
                "is_leaf": node.is_leaf,
                "keys": node.keys,
                "children": []
            }
            if not node.is_leaf:
                for child in node.children:
                    node_data["children"].append(serialize_node(child))
            return node_data

        serialized_tree = serialize_node(bplustree.root)
        serialized_nodes.append(serialized_tree)
        return serialized_nodes

    @staticmethod
    def _deserialize_bplustree(bplustree, data):
        def deserialize_node(node_data):
            node = BPlusTreeNode(is_leaf=node_data["is_leaf"])
            node.keys = node_data["keys"]
            if not node.is_leaf:
                node.children = [deserialize_node(child) for child in node_data["children"]]
            return node

        bplustree.root = deserialize_node(data[0])

    def remove_index(self, field):
        if field in self.indices:
            del self.indices[field]
        if field in self.index_schema:
            del self.index_schema[field]



import lynse
import concurrent.futures

client = lynse.VectorDBClient()


def add_item(id, vector, field):
    db = client.create_database("test_min_vec", drop_if_exists=False)
    collection = db.get_collection('test_collection')
    collection.add_item(vector, id=id, field=field)
    collection.commit()


def bulk_add_items(items):
    db = client.create_database("test_min_vec", drop_if_exists=False)
    collection = db.get_collection('test_collection')
    collection.bulk_add_items(items)
    collection.commit()


def test_multi_thread_bulk_add_items():
    db = client.create_database("test_min_vec", drop_if_exists=False)
    collection = db.require_collection('test_collection', dim=4, drop_if_exists=True)

    items = [
        ([1, 2, 3, 4], 1, {'name': 'John Doe'}),
        ([5, 6, 7, 8], 2, {'name': 'Jane Doe'}),
        ([9, 10, 11, 12], 3, {'name': 'John Smith'}),
        ([13, 14, 15, 16], 4, {'name': 'Jane Smith'})
    ]

    with concurrent.futures.ThreadPoolExecutor() as executor:
        futures = [executor.submit(bulk_add_items, items),
                   executor.submit(bulk_add_items, items),
                   executor.submit(bulk_add_items, items),
                   executor.submit(bulk_add_items, items)]

        concurrent.futures.wait(futures)

    assert collection.shape == (4, 4)
    db.drop_database()



from queue import Queue
import numpy as np


class LimitedSorted:
    def __init__(self, n=10):
        """A class to store the top n most similar vectors to a given vector.

        Parameters:
            n (int): The number of vectors to store.
        """
        self.n = n
        self.similarities = Queue()
        self.indices = Queue()
        self.is_2d = None

    def add(self, sim: np.ndarray, indices: np.ndarray):
        """Add similarities and their corresponding indices.

        Parameters:
            sim (np.ndarray): 1D or 2D array of similarities.
            indices (np.ndarray): 1D or 2D array of corresponding indices.
        """
        if sim.ndim != indices.ndim:
            raise ValueError("sim and indices must have the same number of dimensions.")

        if self.is_2d is None:
            self.is_2d = sim.ndim == 2

        if (sim.ndim == 2) != self.is_2d:
            raise ValueError("All added data must have the same number of dimensions.")

        self.similarities.put(sim)
        self.indices.put(indices)

    def get_top_n(self):
        """Get the top n most similar vectors.

        Returns:
            np.ndarray: The indices of the top n most similar vectors.
            np.ndarray: The similarities of the top n most similar vectors.
        """
        if self.similarities.empty():
            return np.array([]), np.array([])

        if self.is_2d:
            indices = np.vstack(np.atleast_2d(self.indices.queue))
            sim = np.vstack(np.atleast_2d(self.similarities.queue))

            sorted_indices, topk_similarities = self._sort_n_idx_2d(sim, indices)
            return sorted_indices, topk_similarities
        else:
            indices = np.concatenate(np.atleast_2d(self.indices.queue))
            sim = np.concatenate(np.atleast_2d(self.similarities.queue))
            sorted_idx = self._sort_n_idx(sim)
            return indices[sorted_idx], sim[sorted_idx]

    def _sort_n_idx(self, sim):
        """Sort and retrieve indices of the top n similarities for 1D case.

        Parameters:
            sim (np.ndarray): 1D array of similarities.

        Returns:
            np.ndarray: Indices of the top n similarities.
        """
        if sim.shape[0] <= 1000:
            return np.argsort(sim)[:self.n]

        top_n_idx = np.argpartition(sim, self.n - 1)[:self.n]
        sorted_idx = np.argsort(sim[top_n_idx])
        return top_n_idx[sorted_idx]

    def _sort_n_idx_2d(self, sim, indices):
        """Sort and retrieve top n similarities for each row in 2D case using vectorized operations.

        Parameters:
            sim (np.ndarray): 2D array of similarities.
            indices (np.ndarray): 2D array of corresponding indices.

        Returns:
            np.ndarray: Sorted indices of the top n similarities for each row.
            np.ndarray: The top n similarities for each row.
        """
        partitioned_indices = np.argpartition(sim, self.n - 1, axis=1)[:, :self.n]
        top_n_similarities = np.take_along_axis(sim, partitioned_indices, axis=1)

        top_n_indices = np.take_along_axis(indices, partitioned_indices, axis=1)

        sorted_order = np.argsort(top_n_similarities, axis=1)

        top_n_similarities_sorted = np.take_along_axis(top_n_similarities, sorted_order, axis=1)
        top_n_indices_sorted = np.take_along_axis(top_n_indices, sorted_order, axis=1)

        return top_n_indices_sorted, top_n_similarities_sorted

    def __len__(self):
        return len(self.similarities.queue)



from datetime import datetime
from pathlib import Path
import shutil
import time
from typing import Union

import numpy as np
from spinesUtils.asserts import raise_if
from spinesUtils.logging import Logger
from tqdm import tqdm

from ..core_components.locks import ThreadLock
from ..core_components.id_checker import IDChecker
from ..core_components.fields_cache import FieldsCache
from ..storage_layer.storage import PersistentFileStorage
from ..storage_layer.wal import WALStorage


class MatrixSerializer:
    """
    The MatrixSerializer class is used to serialize and deserialize the matrix data.
    """

    def __init__(
            self,
            dim: int,
            collection_path: Union[str, Path],
            logger: Logger,
            chunk_size: int = 100_000,
            dtypes: str = 'float32',
            warm_up: bool = False,
            cache_chunks: int = 20,
    ) -> None:
        """
        Initialize the vector collection.

        Parameters:
            dim (int): Dimension of the vectors.
            collection_path (str): Path to the collections file.
            logger (Logger): The logger object.
            chunk_size (int): The size of each data chunk. Default is 1_000_000.
            dtypes (str): The data type of the vectors. Default is 'float32'.
                Options are 'float16', 'float32' or 'float64'.
            warm_up (bool): Whether to warm up the collection. Default is False.
            cache_chunks (int): The buffer size for the storage worker. Default is 20.

        Raises:
            ValueError: If `chunk_size` is less than or equal to 1.
        """
        self.last_commit_time = None
        # set commit flag, if the flag is True, the collection will not be saved
        self.COMMIT_FLAG = True
        self.IS_DELETED = False

        self.logger = logger

        self.threadlock = ThreadLock()

        # set parent path
        self._initialize_components_path(collection_path)

        self._dtypes_map = {'float16': np.float16, 'float32': np.float32, 'float64': np.float64}

        # set dtypes
        self.dtypes = self._dtypes_map[dtypes]
        # set dim
        self.dim = dim
        # set chunk size
        self.chunk_size = chunk_size

        # initialize the storage worker
        self.storage_worker = PersistentFileStorage(self.collections_path_parent, self.dim,
                                                    self.chunk_size,
                                                    warm_up=warm_up, cache_chunks=cache_chunks)
        self.wal_worker = WALStorage(collection_name=self.collections_path_parent.name,
                                     chunk_size=self.chunk_size, storage_path=self.collections_path_parent,
                                     flush_interval=1)

        self._initialize_fields_index()
        self._initialize_id_checker()

        # log_dir exists and not empty
        if self.wal_worker.log_dir.exists() and any(self.wal_worker.log_dir.iterdir()):
            self.logger.info("Detected uncommitted data, preparing to recover...")
            # replay wal
            self.commit_data(recover=True)

    def _initialize_components_path(self, collections_path):
        """make directory if not exist"""
        self.collections_path_parent = (Path(collections_path).parent.absolute() /
                                        Path(collections_path).absolute().name)

        self.collections_path_parent.mkdir(parents=True, exist_ok=True)

        # sq_model path
        self.sq_model_path = self.collections_path_parent / 'sq_model'
        # field_index path
        self.field_index_path = self.collections_path_parent / 'fields_index'
        # scaled_status path
        self.scaled_status_path = self.collections_path_parent / 'scaled_status.json'
        # set filter path
        self.filter_path = self.collections_path_parent / 'id_filter'

    def _initialize_fields_index(self):
        """initialize fields index"""
        self.field_index = FieldsCache(self.field_index_path)

    def _initialize_id_checker(self):
        """initialize id checker and shape"""
        self.id_filter = IDChecker()

        if self.filter_path.exists():
            self.id_filter.from_file(self.filter_path)
        else:
            if self.collections_path_parent.exists():
                filenames = self.storage_worker.get_all_files()
                for filename in filenames:
                    database, indices = self.storage_worker.read(filename=filename)
                    self.id_filter.add(indices)

        self.last_id = self.id_filter.find_max_value()

    def dataloader(self, filename):
        """
        Generator for loading the database and index.

        Parameters:
            filename (str): The name of the file to load.

        Yields:
            tuple: A tuple of (database, id).

        Raises:
            FileNotFoundError: If the file does not exist.
            IOError: If the file cannot be read.
            PermissionError: If the file cannot be read due to permission issues.
            UnKnownError: If an unknown error occurs.
        """
        return self.storage_worker.read(filename=filename)

    def commit_data(self, recover=False):
        if not recover:
            start_msg = 'Writing chunk to storage...'
            end_msg = 'Writing chunk to storage done.'
        else:
            start_msg = 'Recovering data...'
            end_msg = 'Recovering data done.'

        self.logger.info(start_msg + '\n')

        if self.shape[0] > 0:
            if self.shape[0] % self.chunk_size == 0:
                last_filename = f"chunk_{int(self.storage_worker.get_all_files()[-1].split('_')[-1]) + 1}"
            else:
                last_filename = self.storage_worker.get_all_files()[-1]
        else:
            last_filename = 'chunk_0'

        # Only print progress bar if the logger level is less than 20, which means INFO or DEBUG
        for data, ids, fields in tqdm(self.wal_worker.get_file_iterator(), desc="Data persisting",
                                      total=self.wal_worker.file_number, unit='chunk', disable=self.logger.level > 20):
            self.storage_worker.write(data, ids)
            # store fields index
            for _id, _field in zip(ids, fields):
                self.field_index.store(_field, int(_id))

            # insert data to indexer
            if hasattr(self, "indexer"):
                self.indexer.index_insert(data, ids)

        if hasattr(self, "indexer"):
            if self.indexer.ivf is not None:
                self.indexer.ivf_insert(last_filename)
            # update filenames
            self.indexer.update_filenames()

        self.logger.info(end_msg)
        self.field_index.commit()
        self.wal_worker.reincarnate()

    def commit(self):
        """
        Save the collection, ensuring that all data is written to disk.
        """
        with self.threadlock:
            if not self.COMMIT_FLAG:
                self.logger.info('Saving data...')
                if hasattr(self, 'buffer'):
                    if len(self.buffer) > 0:
                        self.bulk_add_items(self.buffer)

                self.commit_data()

                # save id filter
                self.logger.debug('Saving id filter...')
                self.id_filter.to_file(self.filter_path)

                # remove buffer
                self._remove_buffer()

                self.COMMIT_FLAG = True

                self.last_commit_time = datetime.now()

    def _process_vector_item(self, vector, index, field):
        if index in self.id_filter:
            raise ValueError(f'id {index} already exists.')

        if vector.ndim == 1:
            vector = vector.reshape(1, -1)

        if vector.shape[1] != self.dim:
            raise ValueError(f'vector dim error, expect {self.dim}, got {vector.shape[1]}')

        if vector.dtype != self.dtypes:
            vector = vector.astype(self.dtypes)

        return vector, index, field if field is not None else {}

    def bulk_add_items(self, vectors):
        """
        Bulk add vectors to the collection in batches.

        It is recommended to use incremental ids for best performance.

        Parameters:
            vectors (list or tuple): A list or tuple of vectors to be saved.
                Each vector can be a tuple of (vector, id, field).

        Returns:
            list: A list of indices where the vectors are stored.
        """
        raise_if(ValueError, not isinstance(vectors, (tuple, list)),
                 f'vectors must be tuple or list, got {type(vectors)}')

        data = []
        indices = []
        fields = []

        for i in range(0, len(vectors), self.chunk_size):
            batch = vectors[i:i + self.chunk_size]

            for sample in batch:
                sample_len = len(sample)

                if sample_len == 3:
                    vector, id, field = sample
                elif sample_len == 2:
                    vector, id = sample
                    field = {}
                else:
                    raise ValueError('Each sample must be a tuple of (vector, id, field[optional]).')

                raise_if(TypeError, not (isinstance(field, dict) or field is None),
                         f'field must be dict or None, got {type(field)}')

                if isinstance(vector, list):
                    vector = np.array(vector)

                raise_if(ValueError, (not isinstance(id, int)) or id < 0,
                         f'id must be integer and greater than 0, got {id}')

                field = {} if field is None else field
                vector, id, field = self._process_vector_item(vector, id, field)

                data.append(vector)
                indices.append(id)
                fields.append(field)

                self.id_filter.add(id)

        self.wal_worker.write_log_data(data, indices, fields)
        self.COMMIT_FLAG = False

        return indices

    def _define_buffer(self, buffer_size: int):
        self.buffer = []
        self.buffer_size = buffer_size

    def _remove_buffer(self):
        if hasattr(self, 'buffer'):
            del self.buffer
        if hasattr(self, 'buffer_size'):
            del self.buffer_size

    def _insert_buffer(self, vector, id: int, field: dict, buffer_size: int):
        if not hasattr(self, 'buffer'):
            self._define_buffer(buffer_size)
        elif getattr(self, 'buffer_size') != buffer_size:
            self._remove_buffer()
            self._define_buffer(buffer_size)

        if len(self.buffer) < self.buffer_size:
            self.buffer.append((vector, id, field))
        else:
            self.bulk_add_items(self.buffer)
            self.buffer = []
            self.buffer.append((vector, id, field))

    def add_item(self, vector, id: int, field: dict = None,
                 buffer_size: Union[None, int, bool] = True) -> int:
        """
        Add a single vector to the collection.

        It is recommended to use incremental ids for best performance.

        Parameters:
            vector (np.ndarray): The vector to be added.
            id (int): The ID of the vector.
            field (dict, optional, keyword-only): The field of the vector. Default is None.
                If None, the field will be set to an empty string.
            buffer_size (int or bool or None): The buffer size for the storage worker. Default is True.
                If None, the vector will be directly written to the disk.
                If True, the buffer_size will be set to chunk_size,
                    and the vectors will be written to the disk when the buffer is full.
                If False, the vector will be directly written to the disk.
                If int, when the buffer is full, the vectors will be written to the disk.

        Returns:
            int: The ID of the added vector.

        Raises:
            ValueError: If the vector dimensions don't match or the ID already exists.
        """
        if isinstance(vector, list):
            vector = np.array(vector)

        raise_if(ValueError, vector.ndim != 1, f'vector dim error, expect 1, got {vector.ndim}')
        raise_if(ValueError, field is not None and not isinstance(field, dict),
                 f'field must be dict, got {type(field)}')

        raise_if(ValueError, (not isinstance(id, int)) or id < 0,
                 f'id must be integer and greater than 0, got {id}')

        vector, id, field = self._process_vector_item(vector, id, field)

        if buffer_size is not None and buffer_size is not False:
            if buffer_size is True:
                buffer_size = self.chunk_size

            self._insert_buffer(vector, id, field, buffer_size)
        else:
            self.wal_worker.write_log_data(vector,
                                           id.reshape(1) if isinstance(id, np.ndarray) else np.array([id]), [field])

        self.COMMIT_FLAG = False

        return id

    def delete(self):
        """Delete collection."""
        with self.threadlock:
            if not self.collections_path_parent.exists():
                return None

            # stop wal
            self.wal_worker.stop()

            retries = 3
            delay = 1  # 每次重试的间隔时间（秒）

            for attempt in range(retries):
                try:
                    shutil.rmtree(self.collections_path_parent)
                    self.logger.info(f"Successfully deleted directory {self.collections_path_parent}")
                    break
                except PermissionError as e:
                    if attempt < retries - 1:
                        self.logger.warning(f"Failed to delete directory, attempting {attempt + 1}...")
                        time.sleep(delay)
                    else:
                        self.logger.error(f"Failed to delete directory {self.collections_path_parent}, file may be locked.")
                        raise e
                except Exception as e:
                    self.logger.error(f"Error deleting: {e}")

            self.IS_DELETED = True

            self._initialize_fields_index()
            self._initialize_id_checker()

            # clear cache
            self.storage_worker.clear_cache()

    @property
    def shape(self):
        """
        Get the shape of the collection.
        """
        with self.threadlock:
            return tuple(self.storage_worker.get_shape(read_type='all'))



import uuid
from pathlib import Path
import json
from typing import Union

import numpy as np
from spinesUtils.asserts import raise_if

from ..core_components.limited_dict import LimitedDict
from ..core_components.locks import ThreadLock
from ..utils.utils import safe_mmap_reader


class PersistentFileStorage:
    """The worker class for reading and writing data to the files."""

    def __init__(self, collection_path, dimension, chunk_size, warm_up=False, cache_chunks=20):
        self.collection_path = Path(collection_path)
        self.collection_name = self.collection_path.name
        self.collection_chunk_path = self.collection_path / 'chunk_data'
        self.collection_chunk_indices_path = self.collection_path / 'chunk_ids'

        # # rename the ids folder, the version upgrade has resulted in compatibility modifications
        if (self.collection_path / 'chunk_indices_data').exists():
            (self.collection_path / 'chunk_indices_data').rename(self.collection_chunk_indices_path)

        self.fingerprint_path = self.collection_path / 'fingerprint'

        self.collection_chunk_path.mkdir(parents=True, exist_ok=True)
        self.collection_chunk_indices_path.mkdir(parents=True, exist_ok=True)

        self.dimension = dimension
        self.chunk_size = chunk_size

        self.lock = ThreadLock()

        self.dataloader = DataLoader(dimension, collection_path, cache_chunks=cache_chunks, warm_up=warm_up)
        self.initialize_fingerprint()

        if warm_up:
            self.dataloader.warm_up()

    def initialize_fingerprint(self):
        with self.lock:
            if self.fingerprint_path.exists():
                with open(self.fingerprint_path, 'r') as f:
                    self.fingerprint = f.readlines()[-1].strip()
            else:
                self.fingerprint: Union[str, None] = None

    def file_exists(self):
        """Check if the file exists."""
        return self.dataloader.file_exists()

    def _return_if_in_memory(self, filename):
        return self.dataloader.return_if_in_memory(filename)

    def _write_to_memory(self, filename, data, indices):
        self.dataloader.write_to_memory(filename, data, indices)

    def get_all_files(self, separate=False):
        return self.dataloader.get_all_files(separate=separate)

    def read(self, filename, return_memory=True):
        """Read data from the specified filename if it exists."""
        return self.dataloader.read(filename, return_memory=return_memory)

    def mmap_read(self, filename):
        return self.dataloader.mmap_read(filename)

    def get_last_id(self):
        ids = [int(i.stem.split('_')[-1])
               for i in self.collection_chunk_path.glob('chunk_*')]

        if len(ids) > 0:
            return max(ids)

        return -1

    def update_fingerprint(self):
        with self.lock:
            self.fingerprint = uuid.uuid4().hex
            with open(self.fingerprint_path, 'a') as f:
                f.write(self.fingerprint + '\n')

    @staticmethod
    def _write_to_disk(data, indices, data_path, indices_path, filename):
        with open(data_path / filename, 'wb') as f:
            np.save(f, data)

        with open(indices_path / filename, 'wb') as f:
            np.save(f, indices)

    def _write(self, data, indices):
        with self.lock:
            if not isinstance(data, np.ndarray):
                data = np.vstack(data)

            if not isinstance(indices, np.ndarray):
                indices = np.array(indices)

            collection_subfile_path = self.collection_chunk_path
            collection_indices_path = self.collection_chunk_indices_path
            file_prefix = 'chunk'

            last_file_id = self.get_last_id()

            # read info file to get the shape of the data
            # file shape
            # save the total shape of the data
            if not (self.collection_path / 'info.json').exists():
                total_shape = [0, self.dimension]
                with open(self.collection_path / 'info.json', 'w') as f:
                    json.dump({"total_shape": total_shape}, f)
            else:
                with open(self.collection_path / 'info.json', 'r') as f:
                    total_shape = json.load(f)['total_shape']

            data_shape = len(data)

            # new file
            if total_shape[0] % self.chunk_size == 0 or last_file_id == -1:
                while len(data) != 0:
                    last_file_id = self.get_last_id()

                    temp_data = data[:self.chunk_size]
                    temp_indices = indices[:self.chunk_size]

                    data = data[self.chunk_size:]
                    indices = indices[self.chunk_size:]

                    filename = f'{file_prefix}_{last_file_id + 1}'
                    # save data
                    self._write_to_disk(temp_data, temp_indices, collection_subfile_path, collection_indices_path,
                                        filename)

                    self._write_to_memory(filename, temp_data, temp_indices)
            # append data to the last file
            else:
                already_stack = False
                while len(data) != 0:
                    last_file_id = self.get_last_id()
                    # run once
                    if not already_stack:
                        temp_index = self.chunk_size - (total_shape[0] % self.chunk_size)
                        temp_data = data[:temp_index]
                        temp_indices = indices[:temp_index]

                        data = data[temp_index:]
                        indices = indices[temp_index:]
                        already_stack = True

                        # save data
                        with open(collection_subfile_path / f'{file_prefix}_{last_file_id}', 'rb') as f:
                            old_data = np.load(f)
                        temp_data = np.vstack((old_data, temp_data))
                        # save indices
                        with open(collection_indices_path / f'{file_prefix}_{last_file_id}', 'rb') as f:
                            old_indices = np.load(f)
                        temp_indices = np.concatenate((old_indices, temp_indices))

                        filename = f'{file_prefix}_{last_file_id}'
                        self._write_to_disk(temp_data, temp_indices, collection_subfile_path,
                                            collection_indices_path, filename)

                        self._write_to_memory(filename, temp_data, temp_indices)
                    else:
                        temp_index = min(self.chunk_size, len(data))
                        temp_data = data[:temp_index]
                        temp_indices = indices[:temp_index]

                        data = data[temp_index:]
                        indices = indices[temp_index:]

                        filename = f'{file_prefix}_{last_file_id + 1}'
                        self._write_to_disk(temp_data, temp_indices, collection_subfile_path,
                                            collection_indices_path, filename)

                        self._write_to_memory(filename, temp_data, temp_indices)

            with open(self.collection_path / 'info.json', 'w') as f:
                total_shape[0] += data_shape
                json.dump({"total_shape": total_shape}, f)

    def read_by_id(self, filename, id=None):
        """Read the data from the file as a memory-mapped file."""
        return self.dataloader.read_by_id(filename, id=id)

    def read_by_only_id(self, id: Union[int, list]):
        """
        Read the data from the file by only the id.

        Parameters:
            id (Union[int, list]): The id or list of ids to read.

        Returns:
            Tuple: The data and the indices.
        """
        return self.dataloader.read_by_only_id(id=id)

    def write(self, data=None, indices=None):
        """Write the data to the file."""
        with self.lock:
            self._write(data, indices=indices)

            # update the fingerprint
            self.update_fingerprint()

    def get_shape(self, read_type='all'):
        """Get the shape of the data.
        parameters:
            read_type (str): The type of data to read. Must be 'chunk' or 'all'.
        """
        return self.dataloader.get_shape(read_type=read_type)

    def clear_cache(self):
        self.dataloader.clear_cache()


class DataLoader:
    def __init__(self, dimension, collection_path, cache_chunks=20, warm_up=False):
        self.dimension = dimension
        self.collection_path = Path(collection_path)
        self.collection_chunk_path = self.collection_path / 'chunk_data'
        self.collection_chunk_indices_path = self.collection_path / 'chunk_ids'
        self.cache = LimitedDict(cache_chunks) if (cache_chunks > 0 or cache_chunks == -1) else None
        self.lock = ThreadLock()
        if warm_up:
            self.warm_up()

    def file_exists(self):
        """Check if the file exists."""
        return (self.collection_chunk_path / 'chunk_0').exists()

    def mmap_read(self, filename):
        return safe_mmap_reader(self.collection_chunk_path / filename), \
               safe_mmap_reader(self.collection_chunk_indices_path / filename)

    def warm_up(self):
        """Load the data from the file to the memory."""
        with self.lock:
            if not self.file_exists():
                return

            if self.cache is not None:
                filenames = self.get_all_files()

                for idx, filename in enumerate(filenames):
                    if not self.cache.is_reached_max_size:
                        self.read(filename, return_memory=False)

    def write_to_memory(self, filename, data, indices):
        with self.lock:
            if self.cache is not None:
                if not self.cache.is_reached_max_size:
                    self.cache[filename] = (data, indices)

    def return_if_in_memory(self, filename):
        with self.lock:
            if self.cache is None:
                return None

        res = self.cache.get(filename)

        if res is None or (len(res) == 2 and res[0] is None):
            return None

        return res

    def load_data(self, filename, data_path, indices_path, update_memory=True):
        with self.lock:
            with open(data_path / filename, 'rb') as f:
                data = np.load(f)
            with open(indices_path / filename, 'rb') as f:
                indices = np.load(f)

        if update_memory:
            self.write_to_memory(filename, data, indices)

        return data, indices

    def get_all_files(self, separate=False):
        with self.lock:
            filenames = sorted([x.stem for x in self.collection_chunk_path.glob('chunk_*')],
                               key=lambda x: int(x.split('_')[-1]))
            if separate:
                if self.cache:
                    return [filename for filename in filenames if filename in self.cache], \
                        [filename for filename in filenames if filename not in self.cache]
                return [], filenames

            return filenames

    def read(self, filename, return_memory=True):
        """Read data from the specified filename if it exists."""
        with self.lock:
            if not self.file_exists():
                return

            if not return_memory:
                return self.load_data(filename, self.collection_chunk_path, self.collection_chunk_indices_path)

            return self.return_if_in_memory(filename) or self.load_data(filename, self.collection_chunk_path,
                                                                        self.collection_chunk_indices_path)

    def read_by_id(self, filename, id=None):
        """Read the data from the file as a memory-mapped file."""
        with self.lock:
            idx_filter = lambda x: np.isin(indices, id, assume_unique=True) if id is not None else None

            data, indices = self.read(filename)
            if not isinstance(indices, np.ndarray):
                indices = np.asarray(indices)

            inter_idx = idx_filter(indices)

            if inter_idx is None or (not np.any(inter_idx)):
                return np.array([]), np.array([])

            data = np.asarray(data[inter_idx])
            indices = indices[inter_idx]

            return data, indices

    def read_by_only_id(self, id: Union[int, list]):
        with self.lock:
            if isinstance(id, int):
                id = [id]

            filenames = self.get_all_files()
            data = []
            indices = []
            for filename in filenames:
                _data, _indices = self.read_by_id(filename, id)
                if len(_data) > 0:
                    data.append(_data)
                    indices.append(_indices)

            return np.vstack(data), np.concatenate(indices)

    def get_shape(self, read_type='all'):
        """Get the shape of the data.
        parameters:
            read_type (str): The type of data to read. Must be 'chunk' or 'all'.
        """
        raise_if(ValueError, read_type not in ['chunk', 'all'], 'read_type must be "chunk" or "all".')

        with self.lock:
            if read_type == 'chunk':
                shape = [0, self.dimension]
                filenames = self.get_all_files()
                if filenames:
                    for filename in filenames:
                        data, _ = self.read(filename)
                        shape[0] += len(data)

                return shape

            else:
                if not (self.collection_path / 'info.json').exists():
                    return [self.get_shape('chunk')[0], self.dimension]

                with open(self.collection_path / 'info.json', 'r') as f:
                    return json.load(f)['total_shape']

    def clear_cache(self):
        with self.lock:
            if self.cache is not None:
                self.cache.clear()



from flask import Flask
from waitress import serve
import socket

from lynse.api.http_api.http_api.database_ops import database_ops
from lynse.api.http_api.http_api.collection_ops import collection_ops


app = Flask(__name__)
app.register_blueprint(database_ops)
app.register_blueprint(collection_ops)


def get_local_ip():
    try:
        hostname = socket.gethostname()
        local_ip = socket.gethostbyname(hostname)
        return local_ip
    except Exception as e:
        return "127.0.0.1"


def main():
    import argparse
    parser = argparse.ArgumentParser(description='Start the LynseDB HTTP API server.')
    parser.add_argument('--host', default='127.0.0.1', help='The host to bind to.')
    parser.add_argument('--port', default=7637, type=int, help='The port to bind to.')
    parser.add_argument('--threads', default=10, type=int, help='Number of threads per worker.')
    parser.add_argument('run', help='Run the server.')
    args = parser.parse_args()

    if args.run:
        if args.host == '0.0.0.0':
            local_ip = get_local_ip()
            print(f"Server running at:")
            print(f"  - Localhost: http://localhost:{args.port}")
            print(f"  - Local IP: http://{local_ip}:{args.port}", end="\n\n")
        else:
            print(f"Server running at http://{args.host}:{args.port}", end="\n\n")

        serve(app, host=args.host, port=args.port, threads=args.threads)


def launch_in_jupyter(host: str = '127.0.0.1', port: int = 7637, threads: int = 10):
    """
    Launch the HTTP API server in Python environment.

    Parameters:
        host (str): The host to bind to. Default is '127.0.0.1'
        port (int): The port to bind to. Default is 7637.
        threads (int): Number of threads per worker. Default is 10.

    Returns:
        None
    """
    import threading
    if host == '0.0.0.0':
        local_ip = get_local_ip()
        print(f"Server running at:")
        print(f"  - Localhost: http://localhost:{port}")
        print(f"  - Local IP: http://{local_ip}:{port}", end="\n\n")
    else:
        print(f"Server running at http://{host}:{port}", end="\n\n")

    # set the thread as daemon thread
    server_thread = threading.Thread(target=serve, args=(app,), kwargs={'host': host, 'port': port, 'threads': threads})
    server_thread.daemon = True
    server_thread.start()


if __name__ == '__main__':
    main()



import queue
import time
from datetime import datetime
from typing import Union, List, Tuple

import msgpack
import numpy as np
import httpx
from spinesUtils.asserts import raise_if
from tqdm import trange

from ...core_components.fields_cache import IndexSchema, ExpressionParser
from ...core_components.fields_cache.filter import Filter
from ...api import logger
from ...core_components.locks import ThreadLock
from ...utils.poster import Poster
from ...utils.utils import collection_repr


class ExecutionError(Exception):
    pass


def raise_error_response(response):
    """
    Raise an error response.

    Parameters:
        response: The response from the server.

    Raises:
        ExecutionError: If the server returns an error.
    """
    try:
        rj = response.json()
        raise ExecutionError(rj)
    except Exception as e:
        raise ExecutionError(response.text)


def pack_data(data):
    """
    Pack the data.

    Parameters:
        data: The data to pack.

    Returns:
        bytes: The packed data.
    """
    packed_data = msgpack.packb(data, use_bin_type=True)
    return packed_data


class HTTPClient:
    """
    The HTTPClient class is used to interact with the LynseDB server.
    """
    def __init__(self, uri, database_name):
        """
        Initialize the client.

        Parameters:
            uri (str): The URI of the server, must start with "http://" or "https://".
            database_name (str): The name of the database.

        Raises:
            TypeError: If the URI is not a string.
            ValueError: If the URI does not start with "http://" or "https://".
            ConnectionError: If the server cannot be connected to.
        """

        raise_if(TypeError, not isinstance(uri, str), 'The URI must be a string.')
        raise_if(ValueError, not uri.startswith('http://') or uri.startswith('https://'),
                 'The URI must start with "http://" or "https://".')

        self._session = httpx.Client()

        if uri.endswith('/'):
            self.uri = uri[:-1]
        else:
            self.uri = uri

        self.database_name = database_name

    def require_collection(
            self,
            collection: str,
            dim: int = None,
            chunk_size: int = 100_000,
            dtypes: str = 'float32',
            use_cache: bool = True,
            n_threads: Union[int, None] = 10,
            warm_up: bool = False,
            drop_if_exists: bool = False,
            description: str = None,
            cache_chunks: int = 20
    ):
        """
        Create a collection.

        Parameters:
            collection (str): The name of the collection.
            dim (int): The dimension of the vectors. Default is None.
                When creating a new collection, the dimension of the vectors must be specified.
                When loading an existing collection, the dimension of the vectors is automatically loaded.
            chunk_size (int): The chunk size. Default is 100,000.
            dtypes (str): The data types. Default is 'float32'.
            use_cache (bool): Whether to use cache for search. Default is True.
            n_threads (int): The number of threads. Default is 10.
            warm_up (bool): Whether to warm up. Default is False.
            drop_if_exists (bool): Whether to drop the collection if it exists. Default is False.
            description (str): A description of the collection. Default is None.
                The description is limited to 500 characters.
            cache_chunks (int): The number of chunks to cache in memory. Default is 20.

        Returns:
            Collection: The collection object.

        Raises:
            ConnectionError: If the server cannot be connected to.
        """
        uri = f'{self.uri}/required_collection'

        data = {
            "database_name": self.database_name,
            "collection_name": collection,
            "dim": dim,
            "chunk_size": chunk_size,
            "dtypes": dtypes,
            "use_cache": use_cache,
            "n_threads": n_threads,
            "warm_up": warm_up,
            "drop_if_exists": drop_if_exists,
            "description": description,
            "cache_chunks": cache_chunks
        }

        try:
            response = self._session.post(uri, json=data)
            if response.status_code == 200:
                del data['collection_name']
                del data['database_name']
                collection = Collection(uri=self.uri, database_name=self.database_name,
                                        collection_name=collection, **data)
                return collection
            else:
                raise_error_response(response)
        except httpx.RequestError:
            raise ConnectionError(f'Failed to connect to the server at {uri}.')

    def get_collection(self, collection: str, cache_chunks=20, warm_up=True):
        """
        Get a collection.

        Parameters:
            collection (str): The name of the collection.
            cache_chunks (int): The number of chunks to cache. Default is 20.
            warm_up (bool): Whether to warm up. Default is True.

        Returns:
            Collection: The collection object.

        Raises:
            ExecutionError: If the server returns an error.
        """
        uri = f'{self.uri}/is_collection_exists'
        data = {"database_name": self.database_name, "collection_name": collection}
        response = self._session.post(uri, json=data)

        if response.status_code == 200 and response.json()['params']['exists']:
            uri = f'{self.uri}/get_collection_config'
            data = {"database_name": self.database_name, "collection_name": collection}
            response = self._session.post(uri, json=data)

            params = response.json()['params']['config']
            params.update({'cache_chunks': cache_chunks, 'warm_up': warm_up})

            return Collection(uri=self.uri, database_name=self.database_name, collection_name=collection,
                              **params)
        else:
            raise_error_response(response)

    def drop_collection(self, collection: str):
        """
        Drop a collection.

        Parameters:
            collection (str): The name of the collection.

        Returns:
            dict: The response from the server.

        Raises:
            ExecutionError: If the server returns an error.
        """
        try:
            _ = self.get_collection(collection)
        except ExecutionError:
            pass

        uri = f'{self.uri}/drop_collection'
        data = {"database_name": self.database_name, "collection_name": collection}
        return self._session.post(uri, json=data).json()

    def drop_database(self):
        """
        Drop the database.

        Returns:
            dict: The response from the server.

        Raises:
            ExecutionError: If the server returns an error.
        """
        if not self.database_exists()['params']['exists']:
            return {'status': 'success', 'message': 'The database does not exist.'}

        uri = f'{self.uri}/drop_database'
        data = {"database_name": self.database_name}
        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()
        else:
            raise_error_response(response)

    def database_exists(self):
        """
        Check if the database exists.

        Returns:
            dict: The response from the server.

        Raises:
            ExecutionError: If the server returns an error.
        """
        uri = f'{self.uri}/database_exists'
        data = {"database_name": self.database_name}
        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()
        else:
            raise_error_response(response)

    def show_collections(self):
        """
        Show all collections in the database.

        Returns:
            List: The list of collections.

        Raises:
            ExecutionError: If the server returns an error.
        """
        uri = f'{self.uri}/show_collections'
        data = {"database_name": self.database_name}
        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()['params']['collections']
        else:
            raise_error_response(response)

    def set_environment(self, env: dict):
        """
        Set the environment variables.

        Parameters:
            env (dict): The environment variables. It can be specified on the same time or separately.
                - LYNSE_LOG_LEVEL: The log level.
                - LYNSE_LOG_PATH: The log path.
                - LYNSE_TRUNCATE_LOG: Whether to truncate the log.
                - LYNSE_LOG_WITH_TIME: Whether to log with time.
                - LYNSE_KMEANS_EPOCHS: The number of epochs for KMeans.
                - LYNSE_SEARCH_CACHE_SIZE: The search cache size.
                - LYNSE_DATALOADER_BUFFER_SIZE: The dataloader buffer size.

        Returns:
            dict: The response from the server.

        Raises:
            TypeError: If the value of an environment variable is not a string.
            ExecutionError: If the server returns an error.
        """
        uri = f'{self.uri}/set_environment'

        env_list = ['LYNSE_LOG_LEVEL', 'LYNSE_LOG_PATH', 'LYNSE_TRUNCATE_LOG', 'LYNSE_LOG_WITH_TIME',
                    'LYNSE_KMEANS_EPOCHS', 'LYNSE_SEARCH_CACHE_SIZE', 'LYNSE_DATALOADER_BUFFER_SIZE']

        data = {"database_name": self.database_name}
        for key in env:
            if key in env_list:
                raise_if(TypeError, not isinstance(env[key], str), f'The value of {key} must be a string.')
                data[key] = env[key]

        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()
        else:
            raise_error_response(response)

    def get_environment(self):
        """
        Get the environment variables.

        Returns:
            dict: The response from the server.

        Raises:
            ExecutionError: If the server returns an error.
        """
        uri = f'{self.uri}/get_environment'
        data = {"database_name": self.database_name}
        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()
        else:
            raise_error_response(response)

    def update_collection_description(self, collection: str, description: str):
        """
        Update the description of a collection.

        Parameters:
            collection (str): The name of the collection.
            description (str): The description of the collection.

        Returns:
            dict: The response from the server.

        Raises:
            ExecutionError: If the server returns an error.
        """
        uri = f'{self.uri}/update_collection_description'
        data = {"database_name": self.database_name, "collection_name": collection, "description": description}
        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()
        else:
            raise_error_response(response)

    def show_collections_details(self):
        """
        Show all collections in the database with details.

        Returns:
            pandas.DataFrame: The details of the collections.

        Raises:
            ExecutionError: If the server returns an error.
        """
        uri = f'{self.uri}/show_collections_details'
        data = {"database_name": self.database_name}
        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            rj = response.json()['params']['collections']
            try:
                import pandas as pd
                rj = pd.DataFrame(rj)
            except ImportError:
                ...

            return rj
        else:
            raise_error_response(response)

    def __repr__(self):
        if self.database_exists()['params']['exists']:
            return f"RemoteDatabaseInstance(name={self.database_name}, exists=True)"
        else:
            return f"RemoteDatabaseInstance(name={self.database_name}, exists=False)"

    def __str__(self):
        return self.__repr__()


class Collection:
    """
    The Collection class is used to interact with a collection in the LynseDB.
    """
    name = "Remote"

    def __init__(self, uri, database_name, collection_name, **params):
        """
        Initialize the collection.

        Parameters:
            uri (str): The URI of the server.
            database_name (str): The name of the database.
            collection_name (str): The name of the collection.
            **params: The collection parameters.
                - dim (int): The dimension of the vectors.
                - chunk_size (int): The chunk size.
                - dtypes (str): The data types.
                - use_cache (bool): Whether to use cache.
                - n_threads (int): The number of threads.
                - warm_up (bool): Whether to warm up.
                - drop_if_exists (bool): Whether to drop the collection if it exists.
                - cache_chunks (int): The number of chunks to cache.

        """
        self.IS_DELETED = False
        self._uri = uri
        self._database_name = database_name
        self._collection_name = collection_name
        self._session = Poster()
        self._init_params = params

        self.COMMIT_FLAG = False

        self._mesosphere_list = queue.Queue()
        self._lock = ThreadLock()

    def _get_commit_msg(self):
        """
        Get the commit message.

        Returns:
            str: The last commit time.

        Raises:
            ExecutionError: If the server returns an error.
        """
        uri = f'{self._uri}/get_commit_msg'
        data = {"database_name": self._database_name, "collection_name": self._collection_name}

        response = self._session.post(uri, json=data)

        rj = response.json()
        if response.status_code == 200:
            if (rj['params']['commit_msg'] is None or
                    rj['params']['commit_msg'] == 'No commit message found for this collection'):
                return None
            return rj['params']['commit_msg']['last_commit_time']
        else:
            raise ExecutionError(rj)

    def _update_commit_msg(self, last_commit_time):
        """
        Update the commit message.

        Parameters:
            last_commit_time (str): The last commit time.

        Returns:
            dict: The response from the server.

        Raises:
            ExecutionError: If the server returns an error.
        """
        uri = f'{self._uri}/update_commit_msg'
        data = {
            "database_name": self._database_name,
            "collection_name": self._collection_name,
            "last_commit_time": last_commit_time,
        }

        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()
        else:
            try:
                rj = response.json()
                raise ExecutionError(rj)
            except Exception as e:
                print(e)
                raise ExecutionError(response.text)

    def exists(self):
        """
        Check if the collection exists.

        Returns:
            bool: Whether the collection exists.

        Raises:
            ExecutionError: If the server returns an error.

        """
        uri = f'{self._uri}/is_collection_exists'
        data = {"database_name": self._database_name, "collection_name": self._collection_name}

        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()['params']['exists']
        else:
            raise_error_response(response)

    def add_item(self, vector: Union[list[float], np.ndarray], id: int, *, field: Union[dict, None] = None,
                 buffer_size: int = True):
        """
        Add an item to the collection.
        It is recommended to use incremental ids for best performance.

        Parameters:
            vector (list[float], np.ndarray): The vector of the item.
            id (int): The ID of the item.
            field (dict, optional): The fields of the item.
            buffer_size (int or bool): The buffer size.
                Default is True, which means the default buffer size (1000) will be used.
                If buffer_size is 0, the function will add the item directly.
                If buffer_size is greater than 0, the function will add the item to the buffer.
                If buffer_size is False, the function will add the item directly and not use the buffer.
                If buffer_size is True, the function will add the item to the buffer and use the default buffer size.

        Returns:
            int: The ID of the item.
                If delay_num is greater than 0, and the number of items added is less than delay_num,
                the function will return None. Otherwise, the function will return the IDs of the items added.

        Raises:
            ValueError: If the collection has been deleted or does not exist.
            ExecutionError: If the server returns an error.
        """
        if buffer_size is True:
            buffer_size = 1000
        else:
            if buffer_size is False:
                buffer_size = 0
            else:
                raise_if(ValueError, (not isinstance(buffer_size, int)) or buffer_size < 0,
                         'If buffer_size is not bool, it must be a positive integer.')

        if buffer_size == 0:
            uri = f'{self._uri}/add_item'
            headers = {
                'Content-Type': 'application/msgpack'
            }

            data = {
                "database_name": self._database_name,
                "collection_name": self._collection_name,
                "item": {
                    "vector": vector if isinstance(vector, list) else vector.tolist(),
                    "id": id,
                    "field": field if field is not None else {},
                },
            }

            response = self._session.post(uri, data=None, content=pack_data(data), headers=headers)

            if response.status_code == 200:
                self.COMMIT_FLAG = False
                return response.json()['params']['item']['id']
            else:
                raise_error_response(response)
        else:
            with self._lock:
                self._mesosphere_list.put({
                    "vector": vector if isinstance(vector, list) else vector.tolist(),
                    "id": id,
                    "field": field if field is not None else {},
                })

            if self._mesosphere_list.qsize() >= buffer_size:
                mesosphere_list = list(self._mesosphere_list.queue)

                uri = f'{self._uri}/bulk_add_items'
                headers = {
                    'Content-Type': 'application/msgpack'
                }

                data = {
                    "database_name": self._database_name,
                    "collection_name": self._collection_name,
                    "items": mesosphere_list,
                }

                response = self._session.post(uri, data=None, content=pack_data(data), headers=headers)

                if response.status_code == 200:
                    self.COMMIT_FLAG = False
                    self._mesosphere_list = queue.Queue()
                else:
                    raise_error_response(response)

            return id

    @staticmethod
    def _check_bulk_add_items(vectors):
        items = []
        for vector in vectors:
            raise_if(TypeError, not isinstance(vector, tuple), 'Each item must be a tuple of vector, '
                                                               'ID, and fields(optional).')
            vec_len = len(vector)

            if vec_len == 3:
                v1, v2, v3 = vector
                items.append({
                    "vector": v1.tolist() if isinstance(v1, np.ndarray) else v1,
                    "id": v2,
                    "field": v3,
                })
            elif vec_len == 2:
                v1, v2 = vector
                items.append({
                    "vector": v1.tolist() if isinstance(v1, np.ndarray) else v1,
                    "id": v2,
                    "field": {},
                })
            else:
                raise TypeError('Each item must be a tuple of vector, ID, and fields(optional).')

        return items

    def bulk_add_items(
            self,
            vectors: List[Union[
                Tuple[Union[List, Tuple, np.ndarray], int, dict],
                Tuple[Union[List, Tuple, np.ndarray], int]
            ]],
            batch_size: int = 1000,
            enable_progress_bar: bool = True
    ):
        """
        Add multiple items to the collection.
        It is recommended to use incremental ids for best performance.

        Parameters:
            vectors (List[Tuple[Union[List, Tuple, np.ndarray], int, dict]],
            List[Tuple[Union[List, Tuple, np.ndarray], int]]):
                The list of items to add. Each item is a tuple containing the vector, ID, and fields.
            batch_size (int): The batch size. Default is 1000.
            enable_progress_bar (bool): Whether to enable the progress bar. Default is True.

        Returns:
            dict: The response from the server.

        Raises:
            ValueError: If the collection has been deleted or does not exist.
            TypeError: If the vectors are not in the correct format.
            ExecutionError: If the server returns an error.
        """

        uri = f'{self._uri}/bulk_add_items'
        total_batches = (len(vectors) + batch_size - 1) // batch_size

        ids = []

        if enable_progress_bar:
            iter_obj = trange(total_batches, desc='Adding items', unit='batch')
        else:
            iter_obj = range(total_batches)

        headers = {
            'Content-Type': 'application/msgpack',
        }

        for i in iter_obj:
            start = i * batch_size
            end = (i + 1) * batch_size
            items = vectors[start:end]

            items_after_checking = self._check_bulk_add_items(items)

            data = {
                "database_name": self._database_name,
                "collection_name": self._collection_name,
                "items": items_after_checking,
            }

            response = self._session.post(uri, content=pack_data(data), headers=headers)

            if response.status_code == 200:
                self.COMMIT_FLAG = False
                ids.extend(response.json()['params']['ids'])
            else:
                raise_error_response(response)

        return ids

    def commit(self):
        """
        Commit the changes in the collection.

        Returns:
            dict: The response from the server.

        Raises:
            ExecutionError: If the server returns an error.
        """
        uri = f'{self._uri}/commit'
        data = {"database_name": self._database_name, "collection_name": self._collection_name}

        if not self._mesosphere_list.empty():
            data["items"] = list(self._mesosphere_list.queue)

        self._mesosphere_list = queue.Queue()

        response = self._session.post(uri, content=pack_data(data), headers={'Content-Type': 'application/msgpack'})

        if response.status_code == 202:
            task_id = response.json().get('task_id')
            status_uri = f'{self._uri}/status/{task_id}'

            while True:
                status_response = self._session.get(status_uri)
                status_data = status_response.json()

                if status_response.status_code == 200:
                    logger.info(f'Task status: {status_data}', rewrite_print=True)
                    if status_data['status'] in ['Success', 'Error']:
                        if status_data['status'] == 'Success':
                            self._update_commit_msg(datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
                        else:
                            raise_error_response(status_response)
                        return status_data
                else:
                    raise_error_response(status_response)

                time.sleep(2)
        else:
            raise_error_response(response)

    def is_id_exists(self, id: int):
        """
        Check if an ID exists in the collection.

        Parameters:
            id (int): The ID to check.

        Returns:
            is_id_exists(Bool): Whether the ID exists in the collection.
        """
        uri = f'{self._uri}/is_id_exists'
        data = {"database_name": self._database_name, "collection_name": self._collection_name, "id": id}
        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()['params']['is_id_exists']
        else:
            raise_error_response(response)

    @property
    def max_id(self):
        """
        Get the maximum ID in the collection.

        Returns:
            int: The maximum ID in the collection.
        """
        uri = f'{self._uri}/max_id'
        data = {"database_name": self._database_name, "collection_name": self._collection_name}

        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()['params']['max_id']
        else:
            raise_error_response(response)

    def build_index(self, index_mode: str = 'IVF-FLAT', n_clusters: int = 16):
        """
        Build the index of the collection.

        Parameters:
            index_mode (str): The index mode. Default is 'IVF-FLAT'.
            n_clusters (int): The number of clusters. Default is 16.

        Returns:
            dict: The response from the server.

        Raises:
            ExecutionError: If the server returns an error.
        """
        uri = f'{self._uri}/build_index'
        data = {
            "database_name": self._database_name,
            "collection_name": self._collection_name,
            "index_mode": index_mode,
            "n_clusters": n_clusters
        }

        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()
        else:
            raise_error_response(response)

    def remove_index(self):
        """
        Remove the index of the collection.

        Returns:
            dict: The response from the server.

        Raises:
            ExecutionError: If the server returns an error.
        """
        uri = f'{self._uri}/remove_index'
        data = {"database_name": self._database_name, "collection_name": self._collection_name}
        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()
        else:
            raise_error_response(response)

    def insert_session(self):
        """
        Start an insert session.
        """
        from ...execution_layer.session import DataOpsSession

        return DataOpsSession(self)

    def _search(self, vector, k, search_filter, return_fields=False, **kwargs):
        """
        Search the collection.
        """
        uri = f'{self._uri}/search'

        raise_if(ValueError, not isinstance(search_filter, (Filter, type(None), str)),
                 'search_filter must be Filter or None or FieldExpression string.')

        if isinstance(search_filter, str):
            search_filter = ExpressionParser(search_filter).to_filter()

        if search_filter is not None:
            search_filter = search_filter.to_dict()

        data = {
            "database_name": self._database_name,
            "collection_name": self._collection_name,
            "vector": vector if isinstance(vector, list) else vector.tolist(),
            "k": k,
            "search_filter": search_filter,
            "return_fields": return_fields,
            **kwargs
        }
        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()
        else:
            raise_error_response(response)

    def search(
            self, vector: Union[list[float], np.ndarray], k: int = 10, *,
            search_filter: Union[Filter, None] = None,
            return_fields: bool = False, **kwargs
    ):
        """
        Search the database for the vectors most similar to the given vector.

        Parameters:
            vector (np.ndarray or list): The search vectors, it can be a single vector or a list of vectors.
                The vectors must have the same dimension as the vectors in the database,
                and the type of vector can be a list or a numpy array.
            k (int): The number of nearest vectors to return.
            search_filter (Filter or FilterExpression string, optional): The filter to apply to the search.
            return_fields (bool): Whether to return the fields of the search results.
            kwargs: Additional keyword arguments. The following are valid:
                rescore (bool): Whether to rescore the results of binary or scaler quantization searches.
                    Default is False. It is recommended to set it to True when the index mode is 'Binary'.
                rescore_multiplier (int): The multiplier for the rescore operation.
                    It is only available when rescore is True.
                    If 'Binary' is in the index mode, the default is 10. Otherwise, the default is 2.

        Returns:
            Tuple: If return_fields is True, the indices, similarity scores,
                    and fields of the nearest vectors in the database.
                Otherwise, the indices and similarity scores of the nearest vectors in the database.

        Raises:
            ValueError: If the collection has been deleted or does not exist.
            ExecutionError: If the server returns an error.
        """
        rjson = self._search(vector=vector, k=k, search_filter=search_filter, return_fields=return_fields, **kwargs)

        ids, scores = (np.array(rjson['params']['items']['ids']),
                       np.array(rjson['params']['items']['scores']))
        fields = rjson['params']['items']['fields']

        return ids, scores, fields

    @property
    def shape(self):
        """
        Get the shape of the collection.

        Returns:
            Tuple: The shape of the collection.

        Raises:
            ExecutionError: If the server returns an error.
        """
        uri = f'{self._uri}/collection_shape'
        data = {"database_name": self._database_name, "collection_name": self._collection_name}
        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return tuple(response.json()['params']['shape'])
        else:
            rj = response.json()
            if 'error' in rj and rj['error'] == f"Collection '{self._collection_name}' does not exist.":
                return 0, self._init_params['dim']
            else:
                raise_error_response(response)

    def head(self, n: int = 5):
        """
        Get the first n items in the collection.

        Parameters:
            n (int): The number of items to return. Default is 5.

        Returns:
            Tuple: The vectors, IDs, and fields of the items.

        Raises:
            ExecutionError: If the server returns an error.
        """
        uri = f'{self._uri}/head'
        data = {"database_name": self._database_name, "collection_name": self._collection_name, "n": n}
        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            head = response.json()['params']['head']
            return np.asarray(head[0]), np.asarray(head[1]), head[2]
        else:
            raise_error_response(response)

    def tail(self, n: int = 5):
        """
        Get the last n items in the collection.

        Parameters:
            n (int): The number of items to return. Default is 5.

        Returns:
            Tuple: The vectors, IDs, and fields of the items.

        Raises:
            ExecutionError: If the server returns an error.
        """
        uri = f'{self._uri}/tail'
        data = {"database_name": self._database_name, "collection_name": self._collection_name, "n": n}
        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            tail = response.json()['params']['tail']
            return np.asarray(tail[0]), np.asarray(tail[1]), tail[2]
        else:
            raise_error_response(response)

    def read_by_only_id(self, id: Union[int, list]):
        """
        Read the item by ID.

        Parameters:
            id (int, list): The ID of the item or a list of IDs.

        Returns:
            Tuple: The vectors, IDs, and fields of the items.

        Raises:
            ExecutionError: If the server returns an error.
        """
        uri = f'{self._uri}/read_by_only_id'
        data = {"database_name": self._database_name, "collection_name": self._collection_name, "id": id}
        response = self._session.post(uri, json=data)

        item = response.json()['params']['item']

        if response.status_code == 200:
            return item[0], item[1], item[2]
        else:
            raise_error_response(response)

    def query(self, query_filter, filter_ids=None, return_ids_only=False):
        """
        Query the collection.

        Parameters:
            query_filter (Filter or dict): The filter object.
            filter_ids (list[int]): The list of IDs to filter.
            return_ids_only (bool): Whether to return the IDs only.

        Returns:
            List[dict]: The records. If not return_ids_only, the records will be returned.
            List[int]: The external IDs. If return_ids_only, the external IDs will be returned.

        Raises:
            ExecutionError: If the server returns an error.
        """

        uri = f'{self._uri}/query'

        raise_if(ValueError, not isinstance(query_filter, (Filter, str, type(None), dict)),
                 'query_filter must be Filter or dict or FieldExpression string or None.')

        if isinstance(query_filter, str):
            query_filter = ExpressionParser(query_filter).to_filter()

        if query_filter is not None and isinstance(query_filter, Filter):
            query_filter = query_filter.to_dict()

        data = {
            "database_name": self._database_name,
            "collection_name": self._collection_name,
            "query_filter": query_filter,
            "filter_ids": filter_ids,
            "return_ids_only": return_ids_only
        }

        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()['params']['result']
        else:
            raise_error_response(response)

    def query_vectors(self, query_filter, filter_ids=None):
        """
        Query the vector data by the filter.

        Parameters:
            query_filter (Filter or dict or FieldExpression str or None):
                The filter object or the specify data to filter.
            filter_ids (list[int]):
                The list of external IDs to filter. Default is None.

        Returns:
            (Tuple[List[np.ndarray], List[int], List[Dict]]): The vectors, IDs, and fields of the items.
        """
        uri = f'{self._uri}/query_vectors'

        raise_if(ValueError, not isinstance(query_filter, (Filter, str, type(None), dict)),
                 'query_filter must be Filter or dict or FieldExpression string or None.')

        if isinstance(query_filter, str):
            query_filter = ExpressionParser(query_filter).to_filter()

        if query_filter is not None and isinstance(query_filter, Filter):
            query_filter = query_filter.to_dict()

        data = {
            "database_name": self._database_name,
            "collection_name": self._collection_name,
            "query_filter": query_filter,
            "filter_ids": filter_ids,
        }

        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            result = response.json()['params']['result']
            return np.asarray(result[0]), np.array(result[1]), result[2]
        else:
            raise_error_response(response)

    def build_field_index(self, schema, rebuild_if_exists=False):
        """
        Build the field index of the collection.

        Parameters:
            schema (IndexSchema): The schema of the field index.
            rebuild_if_exists (bool): Whether to rebuild the field index if it exists.

        Returns:
            dict: The response from the server.

        Raises:
            ExecutionError: If the server returns an error.
        """
        if not isinstance(schema, IndexSchema):
            raise TypeError("schema must be an instance of IndexSchema.")

        uri = f'{self._uri}/build_field_index'
        data = {
            "database_name": self._database_name,
            "collection_name": self._collection_name,
            "schema": schema.to_dict(),
            "rebuild_if_exists": rebuild_if_exists
        }

        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()
        else:
            raise_error_response(response)

    def list_field_index(self):
        """
        List the field index of the collection.

        Returns:
            dict: The field index of the collection.

        Raises:
            ExecutionError: If the server returns an error.
        """
        uri = f'{self._uri}/list_field_index'
        data = {"database_name": self._database_name, "collection_name": self._collection_name}
        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()['params']['field_indices']
        else:
            raise_error_response(response)

    def remove_field_index(self, field_name):
        """
        Remove the field index of the collection.

        Parameters:
            field_name (str): The name of the field.

        Returns:
            dict: The response from the server.

        Raises:
            ExecutionError: If the server returns an error.
        """
        uri = f'{self._uri}/remove_field_index'
        data = {
            "database_name": self._database_name,
            "collection_name": self._collection_name,
            "field_name": field_name
        }

        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()
        else:
            raise_error_response(response)

    def remove_all_field_indices(self):
        """
        Remove all the field indices of the collection.

        Returns:
            dict: The response from the server.

        Raises:
            ExecutionError: If the server returns an error.
        """
        uri = f'{self._uri}/remove_all_field_indices'
        data = {"database_name": self._database_name, "collection_name": self._collection_name}
        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()
        else:
            raise_error_response(response)

    def list_fields(self):
        """
        List all fields of a collection.

        Returns:
            dict: The status of the operation.
        """
        uri = f'{self._uri}/list_fields'
        data = {"database_name": self._database_name, "collection_name": self._collection_name}
        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()['params']['fields']
        else:
            raise_error_response(response)

    def update_description(self, description: str):
        """
        Update the description of the collection.

        Parameters:
            description (str): The description of the collection.

        Returns:
            dict: The response from the server.

        Raises:
            ExecutionError: If the server returns an error.
        """
        uri = f'{self._uri}/update_description'
        data = {
            "database_name": self._database_name,
            "collection_name": self._collection_name,
            "description": description
        }

        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()
        else:
            raise_error_response(response)

    def get_collection_path(self):
        """
        Get the path of the database.

        Returns:
            str: The path of the database.
        """
        uri = f'{self._uri}/get_collection_path'
        data = {"database_name": self._database_name, "collection_name": self._collection_name}

        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()['params']['collection_path']
        else:
            raise_error_response(response)

    @property
    def index_mode(self):
        """
        Get the index mode of the collection.

        Returns:
            str: The index mode of the collection.
        """
        uri = f'{self._uri}/index_mode'
        data = {"database_name": self._database_name, "collection_name": self._collection_name}
        response = self._session.post(uri, json=data)

        if response.status_code == 200:
            return response.json()['params']['index_mode']
        else:
            raise_error_response(response)

    def __repr__(self):
        return collection_repr(self)

    def __str__(self):
        return self.__repr__()
